"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const collection_1 = require("./collection");
const exception_1 = require("../core/exception");
const lang_1 = require("../core/lang");
class AbstractList extends collection_1.AbstractCollection {
    checkForRange(index) {
        if (index < 0 || index >= this.size()) {
            throw new exception_1.IndexOutOfBoundsException();
        }
    }
    indexOf(o) {
        let index = 0, it = this.iterator();
        while (it.hasNext()) {
            if (it.next() === o) {
                return index;
            }
            index += 1;
        }
        return -1;
    }
    remove(indexOrObject) {
        return lang_1.CoreUtils.isNumber(indexOrObject)
            ? this.removeByIndex(indexOrObject)
            : super.remove(indexOrObject);
    }
    removeByIndex(index) {
        this.checkForRange(index);
        let cursor = 0, it = this.iterator();
        while (it.hasNext()) {
            it.next();
            if (index === cursor) {
                it.remove();
                return true;
            }
            cursor += 1;
        }
        return false;
    }
    // Like most java api, sub string, sub list, all are reference to the original object and limit range as sub object.
    // It make original object cannot be collected by gc even you don't want use, because sub object still effective.
    // That is the reason of create new one as sub list in here.
    subList(fromIndex, toIndex) {
        let newList = new (Object.getPrototypeOf(this).constructor)(), elements = this.toArray().slice(fromIndex, toIndex);
        elements.forEach(elem => newList.add(elem));
        return newList;
    }
}
exports.AbstractList = AbstractList;
class ArrayList extends AbstractList {
    constructor(c) {
        super();
        this.elements = [];
        if (c && c instanceof collection_1.AbstractCollection) {
            this.addAll(c);
        }
    }
    get(index) {
        this.checkForRange(index);
        return this.elements[index];
    }
    indexOf(o) {
        return this.elements.indexOf(o);
    }
    lastIndexOf(o) {
        let elements = this.elements;
        for (let index = elements.length - 1; index >= 0; index -= 1) {
            if (elements[index] === o) {
                return index;
            }
        }
        return -1;
    }
    set(index, element) {
        this.checkForRange(index);
        let oldVal = this.elements[index];
        this.elements[index] = element;
        return oldVal;
    }
    add(e, index) {
        let elements = this.elements;
        if (lang_1.CoreUtils.isNumber(index)) {
            elements.splice(index, 1, ...[e, elements[index]]);
            return true;
        }
        elements.push(e);
        return true;
    }
    iterator() {
        let mod, index = 0, that = this, elements = this.elements.slice();
        return new class {
            next() {
                if (!this.hasNext()) {
                    throw new exception_1.IllegalStateException();
                }
                mod = index;
                index += 1;
                return elements[mod];
            }
            hasNext() {
                return index < elements.length;
            }
            remove() {
                if (lang_1.CoreUtils.isNone(mod) || index - 1 !== mod) {
                    throw new exception_1.IllegalStateException();
                }
                that.elements.splice(index - 1, 1);
            }
        };
    }
    contains(o) {
        return this.elements.indexOf(o) >= 0;
    }
    isEmpty() {
        return this.elements.length <= 0;
    }
    size() {
        return this.elements.length;
    }
    remove(indexOrObject) {
        if (lang_1.CoreUtils.isNumber(indexOrObject)) {
            this.checkForRange(indexOrObject);
            this.elements.splice(indexOrObject, 1);
            return true;
        }
        return this.elements.some((elem, index, elements) => {
            if (elem === indexOrObject) {
                elements.splice(index, 1);
                return true;
            }
            return false;
        });
    }
    clear() {
        this.elements = [];
    }
}
exports.ArrayList = ArrayList;
class ArrayDeque extends collection_1.AbstractCollection {
    constructor(initCapacity = 16, border = false) {
        super();
        this.head = 0;
        this.tail = 0;
        this.border = false;
        this.full = false;
        this.elements = [];
        if (!lang_1.CoreUtils.isNumber(initCapacity)) {
            throw new TypeError(`Expect type number, but got ${typeof initCapacity}`);
        }
        if (!lang_1.CoreUtils.isBoolean(border)) {
            throw new TypeError(`Expect type boolean, but got ${typeof border}`);
        }
        this.border = border;
        this.capacity = initCapacity > lang_1.MAX_INTEGER
            ? lang_1.MAX_INTEGER
            // force to even number
            : parseInt(((initCapacity & 0b1) === 0 ? initCapacity : initCapacity + 1).toFixed(0));
    }
    doubleCapacity() {
        if (this.head !== this.tail) {
            throw new exception_1.IllegalStateException();
        }
        let elements = this.elements, capacity = this.capacity, length = elements.length, head = this.head, halfPart = length - head, newCapacity = [];
        newCapacity.push(...elements.slice(halfPart));
        newCapacity.push(...elements.slice(0, halfPart));
        this.capacity = capacity > lang_1.MAX_INTEGER ? lang_1.MAX_INTEGER : capacity * 2;
        this.elements = newCapacity;
        this.tail = elements.length;
        this.head = 0;
    }
    add(e, index) {
        if (index) {
            throw new exception_1.UnsupportedOperationException();
        }
        this.addLast(e);
        return true;
    }
    contains(o) {
        return this.elements.indexOf(o) >= 0;
    }
    addFirst(e) {
        if (!this.offerFirst(e)) {
            throw new exception_1.IllegalStateException("Deque full.");
        }
    }
    addLast(e) {
        if (!this.offerLast(e)) {
            throw new exception_1.IllegalStateException("Deque full.");
        }
    }
    offerFirst(e) {
        lang_1.Assert.requireNotNull(e, "Element must not be null.");
        if (this.full) {
            return false;
        }
        this.elements[this.head = (this.head - 1) & (this.capacity - 1)] = e;
        if (this.head === this.tail) {
            this.border ? this.full = true : this.doubleCapacity();
        }
        return true;
    }
    offerLast(e) {
        lang_1.Assert.requireNotNull(e, "Element must not be null.");
        if (this.full) {
            return false;
        }
        let tail = this.tail;
        this.elements[tail] = e;
        this.tail = tail = (tail + 1) & (this.capacity - 1);
        if (this.head === tail) {
            this.border ? this.full = true : this.doubleCapacity();
        }
        return true;
    }
    getFirst() {
        let val = this.elements[this.head];
        if (lang_1.CoreUtils.isNone(val)) {
            throw new exception_1.NoSuchElementException();
        }
        return val;
    }
    getLast() {
        let val = this.elements[(this.tail - 1) & (this.capacity - 1)];
        if (lang_1.CoreUtils.isNone(val)) {
            throw new exception_1.NoSuchElementException();
        }
        return val;
    }
    peekFirst() {
        return this.elements[this.head];
    }
    peekLast() {
        return this.elements[(this.tail - 1) & (this.capacity - 1)];
    }
    pollFirst() {
        let head = this.head, elements = this.elements, elem = elements[head];
        if (lang_1.CoreUtils.isNone(elem)) {
            return null;
        }
        this.full = false;
        elements[head] = null;
        this.head = (head + 1) & (this.capacity - 1);
        return elem;
    }
    pollLast() {
        let tail = (this.tail - 1) & (this.capacity - 1), elements = this.elements, elem = elements[tail];
        if (lang_1.CoreUtils.isNone(elem)) {
            return null;
        }
        this.tail = tail;
        this.full = false;
        elements[tail] = null;
        return elem;
    }
    pop() {
        return this.removeFirst();
    }
    push(e) {
        this.addFirst(e);
    }
    removeFirst() {
        let val = this.pollFirst();
        if (lang_1.CoreUtils.isNone(val)) {
            throw new exception_1.NoSuchElementException();
        }
        return val;
    }
    removeLast() {
        let val = this.pollLast();
        if (lang_1.CoreUtils.isNone(val)) {
            throw new exception_1.NoSuchElementException();
        }
        return val;
    }
    element() {
        return this.getFirst();
    }
    offer(e) {
        return this.offerLast(e);
    }
    peek() {
        return this.peekFirst();
    }
    poll() {
        return this.pollFirst();
    }
    remove(o) {
        if (!o) {
            return this.removeFirst();
        }
        let elements = this.elements, index = elements.indexOf(o);
        if (index <= 0) {
            return false;
        }
        this.delete(index);
        return true;
    }
    clear() {
        this.head = this.tail = 0;
        this.elements = [];
    }
    delete(index) {
        let elements = this.elements, mask = this.capacity - 1, head = this.head, tail = this.tail, 
        // steps that has been taken, usual it must be greater than head
        front = (index - head) & mask, 
        // the steps that will be executed, usual it must be less than tail
        back = (tail - index) & mask, arrayCopy = (src, srcPos, dest, destPos, length = 0) => {
            if (length <= 0) {
                return;
            }
            if (srcPos >= src.length || destPos >= dest.length) {
                throw new exception_1.IndexOutOfBoundsException();
            }
            // let copier: T[] = src.slice();
            if (src === dest) {
                let copier = [];
                // set flag, keep same index with src array
                copier[srcPos - 1] = null;
                src = copier.concat(src.slice(srcPos, srcPos + length));
            }
            for (length -= 1; srcPos < src.length && destPos < dest.length && length >= 0; srcPos += 1, destPos += 1, length -= 1) {
                dest[destPos] = src[srcPos];
            }
        };
        if (front < back) {
            if (head <= index) {
                arrayCopy(elements, head, elements, head + 1, front);
            }
            else {
                arrayCopy(elements, 0, elements, 1, index);
                elements[0] = elements[mask];
                arrayCopy(elements, head, elements, head + 1, mask - head);
            }
            elements[head] = null;
            this.head = (index + 1) & mask;
            return false;
        }
        // back >= front
        if (index < tail) {
            arrayCopy(elements, index + 1, elements, index, back);
            this.tail = tail - 1;
        }
        else {
            arrayCopy(elements, index + 1, elements, index, mask - index);
            elements[mask] = elements[0];
            arrayCopy(elements, 1, elements, 0, tail);
            this.tail = (tail - 1) & mask;
        }
        return true;
    }
    iterator() {
        let mod, that = this, mask = this.capacity - 1, elements = this.elements.slice(), last, head = this.head, tail = this.tail;
        let avoid = 0;
        return new class {
            next() {
                if (!this.hasNext()) {
                    return null;
                }
                let elem = elements[head];
                mod = 1;
                last = head;
                head = (head + 1) & mask;
                return elem;
            }
            hasNext() {
                return head !== tail && ((avoid += 1) < 1000);
            }
            remove() {
                if (mod !== 1) {
                    throw new exception_1.IllegalStateException();
                }
                if (that.delete(last)) {
                    head = (head - 1) & mask;
                    tail = that.tail;
                }
                elements = that.elements;
                mod = -1;
            }
        };
    }
    isEmpty() {
        return this.head === this.tail;
    }
    size() {
        return (this.tail - this.head) & (this.capacity - 1);
    }
}
exports.ArrayDeque = ArrayDeque;
class LinkedList extends AbstractList {
    constructor(c) {
        super();
        this.length = 0;
        this.head = new LinkedList.Node(null, null, null);
        this.tail = new LinkedList.Node(null, null, null);
        this.clear();
        if (c && !collection_1.isCollection(c)) {
            this.addAll(c);
        }
    }
    get(index) {
        let cursor = 0, head = this.head, tail = this.tail;
        while ((head = head.next) !== tail) {
            if (index === cursor) {
                return head.val;
            }
            cursor += 1;
        }
        return null;
    }
    lastIndexOf(o) {
        if (this.length <= 0) {
            return -1;
        }
        let index = this.length - 1, head = this.head, tail = this.tail;
        while ((tail = tail.prev) !== head) {
            if (tail.val === o) {
                return index;
            }
            index -= 1;
        }
        return -1;
    }
    set(index, element) {
        this.checkForRange(index);
        let cursor = 0, head = this.head, tail = this.tail;
        while ((head = head.next) !== tail) {
            if (cursor === index) {
                break;
            }
            cursor += 1;
        }
        let oldVal = head.val;
        head.val = element;
        return oldVal;
    }
    add(e, index) {
        let head = this.head, tail = this.tail, node = new LinkedList.Node(e, null, null);
        if (lang_1.CoreUtils.isNumber(index) && !this.checkForRange(index)) {
            let cursor = this.size() - 1;
            while ((tail = tail.prev) !== head) {
                if (cursor === index) {
                    break;
                }
                cursor -= 1;
            }
        }
        node.next = tail;
        node.prev = tail.prev;
        head.next = tail.prev = node;
        this.length += 1;
        return true;
    }
    removeNode(n) {
        n.next.prev = n.prev;
        n.prev.next = n.next;
        n.next = n.prev = null;
        this.length -= 1;
    }
    iterator() {
        let that = this, mod = null, cursor = this.head, tail = this.tail;
        return new class {
            next() {
                if ((mod = cursor = cursor.next) === tail) {
                    return (mod = null);
                }
                return cursor.val;
            }
            hasNext() {
                return cursor !== tail && cursor.next !== tail;
            }
            remove() {
                if (null === mod) {
                    throw new exception_1.IllegalStateException();
                }
                let tmp = cursor.prev;
                that.removeNode(cursor);
                cursor = tmp;
                mod = null;
            }
        };
    }
    size() {
        return this.length;
    }
    addFirst(e) {
        if (!this.offerFirst(e)) {
            throw new exception_1.IllegalStateException("Deque full.");
        }
    }
    addLast(e) {
        if (!this.offerLast(e)) {
            throw new exception_1.IllegalStateException("Deque full.");
        }
    }
    getFirst() {
        let node = this.head.next;
        if (node === this.tail) {
            throw new exception_1.NoSuchElementException();
        }
        return node.val;
    }
    getLast() {
        let node = this.tail.prev;
        if (node === this.head) {
            throw new exception_1.NoSuchElementException();
        }
        return node.val;
    }
    offerFirst(e) {
        let head = this.head, node = new LinkedList.Node(e, null, null);
        node.prev = head;
        node.next = head.next;
        head.next = head.next.prev = node;
        this.length += 1;
        return true;
    }
    offerLast(e) {
        let tail = this.tail, node = new LinkedList.Node(e, null, null);
        node.next = tail;
        node.prev = tail.prev;
        tail.prev = tail.prev.next = node;
        this.length += 1;
        return true;
    }
    peekFirst() {
        let node = this.head.next;
        return node !== this.tail ? node.val : null;
    }
    peekLast() {
        let node = this.tail.prev;
        return node !== this.head ? node.val : null;
    }
    pollFirst() {
        let node = this.head.next;
        if (node === this.tail) {
            return null;
        }
        this.removeNode(node);
        return node.val;
    }
    pollLast() {
        let node = this.tail.prev;
        if (node === this.head) {
            return null;
        }
        this.removeNode(node);
        return node.val;
    }
    pop() {
        return this.removeFirst();
    }
    push(e) {
        this.addFirst(e);
    }
    removeFirst() {
        let val = this.pollFirst();
        if (lang_1.CoreUtils.isNone(val)) {
            throw new exception_1.NoSuchElementException();
        }
        return val;
    }
    removeLast() {
        let val = this.pollLast();
        if (lang_1.CoreUtils.isNone(val)) {
            throw new exception_1.NoSuchElementException();
        }
        return val;
    }
    element() {
        return this.getFirst();
    }
    offer(e) {
        return this.offerLast(e);
    }
    peek() {
        return this.peekFirst();
    }
    poll() {
        return this.pollFirst();
    }
    remove(o) {
        if (!o) {
            return this.removeFirst();
        }
        let head = this.head, tail = this.tail;
        while ((head = head.next) !== tail) {
            if (head.val === o) {
                this.removeNode(head);
                return true;
            }
        }
        return false;
    }
    clear() {
        this.head.next = this.tail;
        this.tail.prev = this.head;
        this.length = 0;
    }
}
LinkedList.Node = class _Node {
    constructor(val, next, prev) {
        this.val = val;
        this.next = next;
        this.prev = prev;
    }
};
exports.LinkedList = LinkedList;
//# sourceMappingURL=list.js.map