"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const collection_1 = require("./collection");
const lang_1 = require("../core/lang");
const exception_1 = require("../core/exception");
function isMap(m) {
    if (!(m instanceof AbstractMap)) {
        throw new TypeError(`Expect type Map, but got ${typeof m}`);
    }
}
class AbstractMap {
    size() {
        return this.entrySet().size();
    }
    clear() {
        this.entrySet().clear();
    }
    containsKey(key) {
        for (let k of this.keySet()) {
            if (k === key) {
                return true;
            }
        }
        return false;
    }
    containsValue(value) {
        for (let v of this.values()) {
            if (v === value) {
                return true;
            }
        }
        return false;
    }
    get(key) {
        for (let entry of this.entrySet()) {
            if (entry.key === key) {
                return entry.value;
            }
        }
        return null;
    }
    getOrDefault(key, defaultVal) {
        return this.containsKey(key) ? this.get(key) : defaultVal;
    }
    isEmpty() {
        return this.size() <= 0;
    }
    keySet() {
        let that = this;
        return new class extends AbstractSet {
            iterator() {
                let it = that.entrySet().iterator();
                return new class {
                    next() {
                        return it.next().key;
                    }
                    hasNext() {
                        return it.hasNext();
                    }
                    remove() {
                        it.remove();
                    }
                };
            }
            size() {
                return that.size();
            }
            contains(o) {
                return that.containsKey(o);
            }
            isEmpty() {
                return that.isEmpty();
            }
            clear() {
                that.clear();
            }
        };
    }
    putAll(map) {
        lang_1.Assert.requireNotNull(map);
        isMap(map);
        for (let entry of map.entrySet()) {
            this.put(entry.key, entry.value);
        }
    }
    remove(key) {
        let entry, it = this.entrySet().iterator();
        while (it.hasNext()) {
            entry = it.next();
            if (entry.key === key) {
                it.remove();
                return entry.value;
            }
        }
        return null;
    }
    values() {
        let that = this;
        return new class extends collection_1.AbstractCollection {
            iterator() {
                let it = that.entrySet().iterator();
                return new class {
                    next() {
                        return it.next().value;
                    }
                    hasNext() {
                        return it.hasNext();
                    }
                    remove() {
                        it.remove();
                    }
                };
            }
            size() {
                return that.size();
            }
            contains(o) {
                return that.containsValue(o);
            }
            clear() {
                that.clear();
            }
        };
    }
    toString() {
        let entries = Object.create(null);
        for (let entry of this.entrySet()) {
            entries[entry.key] = entry.value;
        }
        return JSON.stringify(entries);
    }
}
exports.AbstractMap = AbstractMap;
class AbstractSet extends collection_1.AbstractCollection {
    add(e) {
        throw new exception_1.UnsupportedOperationException();
    }
    addAll(c) {
        lang_1.Assert.requireNotNull(c);
        collection_1.isCollection(c);
        for (let item of c) {
            this.add(item);
        }
        return true;
    }
    remove(o) {
        return super.remove(o);
    }
}
exports.AbstractSet = AbstractSet;
class HashMap extends AbstractMap {
    constructor(initCapacity = 16, loadFactor = .75) {
        super();
        this.elements = [];
        this.length = 0;
        this.loadFactor = loadFactor;
        this.capacity = 16 === initCapacity ? initCapacity : this.calculateSize(initCapacity);
        this.threshold = initCapacity * loadFactor;
    }
    static hash(key) {
        let h;
        return lang_1.CoreUtils.isNone(key) ? 0 : (h = key.hashCode) & (h >>> 16);
    }
    resize() {
        if (this.length < this.threshold) {
            return;
        }
        this.reload(this.capacity * 2);
    }
    reload(capacity) {
        capacity = capacity > lang_1.MAX_INTEGER ? lang_1.MAX_INTEGER : capacity;
        let elem, cursor, newCapacity = [], oldCapacity = this.elements, threshold = capacity * this.loadFactor;
        for (let index = 0, size = oldCapacity.length; index < size; index += 1) {
            elem = oldCapacity[index];
            while (!lang_1.CoreUtils.isNone(elem)) {
                cursor = elem.hash & (capacity - 1);
                if (!newCapacity[cursor]) {
                    newCapacity[cursor] = new HashMap.Entry(elem.key, elem.value, elem.hash, null);
                }
                else {
                    let node = newCapacity[cursor];
                    while (node.next !== null) {
                        node = node.next;
                    }
                    node.next = new HashMap.Entry(elem.key, elem.value, elem.hash, null);
                }
                elem = elem.next;
            }
        }
        this.capacity = capacity;
        this.threshold = threshold;
        this.elements = newCapacity;
    }
    calculateSize(addSize) {
        let size = 2, requireCap = Math.min((this.length + addSize) / this.loadFactor, lang_1.MAX_INTEGER);
        // use '>>>' to remove sign flag
        while (size < requireCap) {
            size = (size <<= 1) >>> 0;
        }
        // cannot less than sixteen
        return size < 16 ? 16 : size;
    }
    getIterator() {
        let node, mod, index = 0, cursor = 0, that = this, elements = this.elements;
        return new class {
            next() {
                if (!this.hasNext()) {
                    return null;
                }
                if (!lang_1.CoreUtils.isNone(node) && lang_1.CoreUtils.isNone(node.next)) {
                    node = mod = null;
                }
                // find node from other bucket if linked list is done
                while (lang_1.CoreUtils.isNone(node) && index < elements.length) {
                    if (!lang_1.CoreUtils.isNone(node = mod = elements[index])) {
                        index += 1;
                        cursor += 1;
                        return node;
                    }
                    index += 1;
                }
                cursor += 1;
                // find node from linked list
                return node = mod = node.next;
            }
            hasNext() {
                return cursor < that.length;
            }
            remove() {
                if (mod !== node || lang_1.CoreUtils.isNone(node)) {
                    throw new exception_1.IllegalStateException();
                }
                let head, prevIndex = index - 1;
                while ((head = elements[prevIndex]) === null) {
                    prevIndex -= 1;
                }
                // back a step if node is head
                if (node === head) {
                    elements[prevIndex] = node.next;
                    node.next = null;
                    index = prevIndex;
                }
                else {
                    while (head && head.next !== node) {
                        head = head.next;
                    }
                    if (!head) {
                        return;
                    }
                    head.next = node.next;
                    node = node.next = null;
                }
                cursor -= 1;
                that.length -= 1;
            }
        };
    }
    entrySet() {
        let that = this;
        return new class extends AbstractSet {
            iterator() {
                let it = that.getIterator();
                return new class {
                    next() {
                        let entry = it.next();
                        return new HashMap.Entry(entry.key, entry.value, -1, null);
                    }
                    hasNext() {
                        return it.hasNext();
                    }
                    remove() {
                        it.remove();
                    }
                };
            }
            size() {
                return that.size();
            }
            contains(o) {
                return that.containsKey(o);
            }
            remove(o) {
                that.remove(o);
                return true;
            }
            clear() {
                that.clear();
            }
        };
    }
    put(key, value) {
        let elements = this.elements, hash = HashMap.hash(key), node = elements[hash & (this.capacity - 1)];
        if (lang_1.CoreUtils.isNone(node)) {
            elements[hash & (this.capacity - 1)] = new HashMap.Entry(key, value, hash, null);
        }
        else {
            let last = node;
            while (node !== null) {
                if (node.hash === hash && (key === node.key || lang_1.CoreUtils.eq(key, node.key))) {
                    let oldVal = node.value;
                    node.value = value;
                    return oldVal;
                }
                last = node;
                node = node.next;
            }
            last.next = new HashMap.Entry(key, value, hash, null);
        }
        this.length += 1;
        this.resize();
        return null;
    }
    putAll(map) {
        lang_1.Assert.requireNotNull(map);
        isMap(map);
        if (this.length + map.size() >= this.threshold) {
            this.reload(this.calculateSize(map.size()));
        }
        for (let entry of map.entrySet()) {
            this.put(entry.key, entry.value);
        }
    }
    remove(key) {
        let elements = this.elements, hash = HashMap.hash(key), index = hash & (this.capacity - 1), node = elements[index];
        if (lang_1.CoreUtils.isNone(node)) {
            return null;
        }
        if (node.hash === hash && (key === node.key || lang_1.CoreUtils.eq(key, node.key))) {
            // one node bucket
            if (node === elements[index] && null === node.next) {
                elements[index] = null;
            }
            else {
                // multi node bucket
                elements[index] = node.next;
                node.next = null;
            }
            this.length -= 1;
            return node.value;
        }
        let last;
        while (node.next !== null) {
            last = node;
            node = node.next;
            if (node.hash === hash && (key === node.key || lang_1.CoreUtils.eq(key, node.key))) {
                last.next = node.next;
                this.length -= 1;
                return node.value;
            }
        }
        return null;
    }
    containsKey(key) {
        let hash = HashMap.hash(key), node = this.elements[hash & (this.capacity - 1)];
        while (!lang_1.CoreUtils.isNone(node)) {
            if (node.hash === hash && (key === node.key || lang_1.CoreUtils.eq(key, node.key))) {
                return true;
            }
            node = node.next;
        }
        return false;
    }
    containsValue(value) {
        let node, elements = this.elements;
        for (let index = 0, size = elements.length; index < size; index += 1) {
            node = elements[index];
            while (!lang_1.CoreUtils.isNone(node)) {
                if (typeof node.value !== typeof value) {
                    node = node.next;
                    continue;
                }
                if (node.value === value || lang_1.CoreUtils.eq(node.value, value)) {
                    return true;
                }
                node = node.next;
            }
        }
        return false;
    }
    size() {
        return this.length;
    }
    get(key) {
        let hash = HashMap.hash(key), node = this.elements[hash & (this.capacity - 1)];
        while (!lang_1.CoreUtils.isNone(node)) {
            if (node.hash === hash && (key === node.key || lang_1.CoreUtils.eq(key, node.key))) {
                return node.value;
            }
            node = node.next;
        }
        return null;
    }
    values() {
        let that = this;
        return new class extends collection_1.AbstractCollection {
            iterator() {
                let it = that.getIterator();
                return new class {
                    next() {
                        return it.next().value;
                    }
                    hasNext() {
                        return it.hasNext();
                    }
                    remove() {
                        it.remove();
                    }
                };
            }
            size() {
                return that.length;
            }
            clear() {
                that.clear();
            }
            contains(o) {
                return that.containsValue(o);
            }
        };
    }
    keySet() {
        let that = this;
        return new class extends AbstractSet {
            iterator() {
                let it = that.getIterator();
                return new class {
                    next() {
                        return it.next().key;
                    }
                    hasNext() {
                        return it.hasNext();
                    }
                    remove() {
                        it.remove();
                    }
                };
            }
            size() {
                return that.length;
            }
            contains(o) {
                return that.containsKey(o);
            }
            clear() {
                that.clear();
            }
        };
    }
    clear() {
        this.elements = [];
        this.length = 0;
    }
    isEmpty() {
        return this.length <= 0;
    }
}
HashMap.Entry = class {
    constructor(key, value, hash, next) {
        this.key = key;
        this.value = value;
        this.hash = hash;
        this.next = next;
    }
};
exports.HashMap = HashMap;
class HashSet extends AbstractSet {
    constructor() {
        super(...arguments);
        this.map = new HashMap();
    }
    iterator() {
        return this.map.keySet().iterator();
    }
    size() {
        return this.map.size();
    }
    add(e) {
        this.map.put(e, null);
        return true;
    }
    remove(o) {
        this.map.remove(o);
        return true;
    }
    contains(o) {
        return this.map.containsKey(o);
    }
    isEmpty() {
        return this.map.isEmpty();
    }
    clear() {
        this.map.clear();
    }
}
exports.HashSet = HashSet;
//# sourceMappingURL=map.js.map